package org.papervision3d.core.proto
{
    import org.papervision3d.core.*;
    import org.papervision3d.core.math.*;
    import org.papervision3d.core.proto.*;
    
    import flash.display.Sprite;
    import flash.utils.Dictionary;
    
    // The DisplayObject class represents instances of 3D objects that are contained in the scene.
    // That includes all objects in the scene, not only those that can be rendered, but also the camera and its target.
    // The Object3D class supports basic functionality like the x, y and z position of an object, as well as rotationX, rotationY, rotationZ, scaleX, scaleY and scaleZ and visible. It also supports more advanced properties of the object such as its transform Matrix3D.
    // Object3D is not an abstract base class; therefore, you can call Object3D directly. Invoking new DisplayObject() creates a new empty object in 3D space, like when you used createEmptyMovieClip().
    public class Object3D
    {
        // An Number that sets the X coordinate of a object relative to the scene coordinate system.
        public function get x():Number
        {
            return this.transform.n14;
        }
    
        public function set x(value:Number):void
        {
            this.transform.n14 = value;
        }
    
        // An Number that sets the Y coordinate of a object relative to the scene coordinates.
        public function get y():Number
        {
            return this.transform.n24;
        }
    
        public function set y(value:Number):void
        {
            this.transform.n24 = value;
        }
    
        // An Number that sets the Z coordinate of a object relative to the scene coordinates.
        public function get z():Number
        {
            return this.transform.n34;
        }
    
        public function set z(value:Number):void
        {
            this.transform.n34 = value;
        }
    
        // Specifies the rotation around the X axis from its original orientation.
        public function get rotationX():Number
        {
            if (this._rotationDirty) 
                updateRotation();
    
            return -this._rotationX * toDEGREES;
        }
    
        public function set rotationX(rot:Number):void
        {
            this._rotationX = -rot * toRADIANS;
            this._transformDirty = true;
        }
    
        // Specifies the rotation around the Y axis from its original orientation.
        public function get rotationY():Number
        {
            if (this._rotationDirty) 
                updateRotation();
    
            return -this._rotationY * toDEGREES;
        }
    
        public function set rotationY(rot:Number):void
        {
            this._rotationY = -rot * toRADIANS;
            this._transformDirty = true;
        }
    
        // Specifies the rotation around the Z axis from its original orientation.
        public function get rotationZ():Number
        {
            if (this._rotationDirty) 
                updateRotation();
    
            return -this._rotationZ * toDEGREES;
        }
    
        public function set rotationZ(rot:Number):void
        {
            this._rotationZ = -rot * toRADIANS;
            this._transformDirty = true;
        }
    
        // Update rotation values
        private function updateRotation():void
        {
            var rot:Number3D = Matrix3D.matrix2euler(this.transform);
            this._rotationX = rot.x * toRADIANS;
            this._rotationY = rot.y * toRADIANS;
            this._rotationZ = rot.z * toRADIANS;
    
            this._rotationDirty = false;
        }
    
        public function set scale(scale:Number):void
        {
            this._scaleX = this._scaleY = this._scaleZ = scale;
    
            this._transformDirty = true;
        }
    
        public function set scaleX(scale:Number):void
        {
            this._scaleX = scale;
    
            this._transformDirty = true;
        }
    
        public function set scaleY(scale:Number):void
        {
            this._scaleY = scale;
    
            this._transformDirty = true;
        }
    
        public function set scaleZ(scale:Number):void
        {
            this._scaleZ = scale;
    
            this._transformDirty = true;
        }
    
        public function get position():Number3D
        {
            return new Number3D(x, y, z);
        }
        // Whether or not the display object is visible.
        // A Boolean value that indicates whether the object is projected, transformed and rendered. A value of false will effectively ignore the object. The default value is true.
        public var visible:Boolean = true;
    
        // An optional object name.
        public var name:String;
    
        // An object that contains user defined properties.
        public var extra:Object; // = {}; TBD
    
        // A Matrix3D object containing values that affect the scaling, rotation, and translation of the display object.
        public var transform:Matrix3D = new Matrix3D();

        private var _transformDirty:Boolean = false;
        private var _rotationDirty:Boolean = false;
    
        private var _rotationX:Number;
        private var _rotationY:Number;
        private var _rotationZ:Number;
    
        private var _scaleX:Number;
        private var _scaleY:Number;
        private var _scaleZ:Number;

    
        // Creates a new Object3D instance. After creating the instance, call the addChild() method of a ObjectContainer3D.
        // @param name        [optional] - The name of the newly created object.
        // @param init  [optional] - An object that contains user defined properties with which to populate the newly created Object3D.
        // x: An Number that sets the X coordinate of a object relative to the scene coordinate system.
        // y: An Number that sets the Y coordinate of a object relative to the scene coordinate system.
        // z: An Number that sets the Z coordinate of a object relative to the scene coordinate system.
        // rotationX: Specifies the rotation around the X axis from its original orientation.
        // rotationY: Specifies the rotation around the Y axis from its original orientation.
        // rotationZ: Specifies the rotation around the Z axis from its original orientation.
        // scaleX: Sets the scale along the local X axis as applied from the registration point of the object.
        // scaleY: Sets the scale along the local Y axis as applied from the registration point of the object.
        // scaleZ: Sets the scale along the local Z axis as applied from the registration point of the object.
        // visible: Whether or not the display object is visible.
        // A Boolean value that indicates whether the object is projected, transformed and rendered. A value of false will effectively ignore the object. The default value is true.
        // container: The MovieClip that you draw into when rendering. Use only when the object is rendered in its own unique MovieClip.
        // extra: An object that contains user defined properties.
        // All properties of the extra field are copied into the new instance. The properties specified with extra are publicly available.
        public function Object3D(name:String = null, init:Object = null):void
        {
            this.name = name;
    
            x = init ? init.x || 0 : 0;
            y = init ? init.y || 0 : 0;
            z = init ? init.z || 0 : 0;
    
            rotationX = init ? init.rotationX || 0 : 0;
            rotationY = init ? init.rotationY || 0 : 0;
            rotationZ = init ? init.rotationZ || 0 : 0;
    
            scaleX = init ? init.scaleX || 1 : 1;
            scaleY = init ? init.scaleY || 1 : 1;
            scaleZ = init ? init.scaleZ || 1 : 1;
    
            extra = init ? init.extra || null : null;

            visible = true;
        }
    
        // Gets the distance to the position of the given object.
        // @param obj The display object to measure the distance to.
        // @return The distance to the registration point of the given object.
        public function distanceTo(obj:Object3D):Number
        {
            var x:Number = this.x - obj.x;
            var y:Number = this.y - obj.y;
            var z:Number = this.z - obj.z;
    
            return Math.sqrt(x*x + y*y + z*z);
        }
    
        public function traverse(traverser:Traverser):void
        {
            traverser.apply(this);
        }

        public function project(parentview:Matrix3D):Matrix3D
        {
            updateTransform();

            return Matrix3D.multiply(parentview, transform);
        }       
    
        // Translate the display object in the direction it is facing, i.e. it's positive Z axis.
        // @param    distance    The distance that the object should move forward.
        public function moveForward(distance:Number):void 
        { 
            translate(distance, Number3D.FORWARD); 
        }
    
        // Translate the display object in the opposite direction it is facing, i.e. it's negative Z axis.
        // @param    distance    The distance that the object should move backward.
        public function moveBackward(distance:Number):void 
        { 
            translate(distance, Number3D.BACKWARD); 
        }
    
        // Translate the display object lateraly, to the left of the direction it is facing, i.e. it's negative X axis.
        // @param distance The distance that the object should move left.
        public function moveLeft(distance:Number):void 
        { 
            translate(distance, Number3D.LEFT); 
        }
    
        // Translate the display object lateraly, to the right of the direction it is facing, i.e. it's positive X axis.
        // @param distance The distance that the object should move right.
        public function moveRight(distance:Number):void 
        { 
            translate(distance, Number3D.RIGHT); 
        }
    
        // Translate the display object upwards, with respect to the direction it is facing, i.e. it's positive Y axis.
        // @param distance The distance that the object should move up.
        public function moveUp(distance:Number):void 
        { 
            translate(distance, Number3D.UP); 
        }
    
        // Translate the display object downwards, with respect to the direction it is facing, i.e. it's negative Y axis.
        // @param distance The distance that the object should move down.
        public function moveDown(distance:Number):void 
        { 
            translate(distance, Number3D.DOWN); 
        }
    
        // Move the object along a given direction.
        // @param distance The distance that the object should travel.
        // @param axis The direction that the object should move towards.
        public function translate(distance:Number, axis:Number3D):void
        {
            updateTransform();

            var vector:Number3D = axis.clone();
    
            Matrix3D.rotateAxis(transform, vector)
    
            this.x += distance * vector.x;
            this.y += distance * vector.y;
            this.z += distance * vector.z;
        }
    
        // Rotate the display object around its lateral or transverse axis  an axis running from the pilot's left to right in piloted aircraft, and parallel to the wings of a winged aircraft; thus the nose pitches up and the tail down, or vice-versa.
        public function pitch(angle:Number):void
        {
            angle = angle * toRADIANS;
    
            var vector:Number3D = Number3D.RIGHT.clone();
    
            updateTransform();
    
            Matrix3D.rotateAxis(transform, vector);
            var m:Matrix3D = Matrix3D.rotationMatrix(vector.x, vector.y, vector.z, angle);
    
            this.transform.copy3x3(Matrix3D.multiply3x3(m, transform));
    
            this._rotationDirty = true;
        }

        // Rotate the display object around about the vertical axis - an axis drawn from top to bottom.
        public function yaw(angle:Number):void
        {
            angle = angle * toRADIANS;
    
            var vector:Number3D = Number3D.UP.clone();
    
            updateTransform();
    
            Matrix3D.rotateAxis(transform, vector);
            var m:Matrix3D = Matrix3D.rotationMatrix(vector.x, vector.y, vector.z, angle);
    
            this.transform.copy3x3(Matrix3D.multiply3x3(m, transform));
    
            this._rotationDirty = true;
        }
    
        // Rotate the display object around the longitudinal axis - an axis drawn through the body of the vehicle from tail to nose in the normal direction of flight, or the direction the object is facing.
        public function roll(angle:Number):void
        {
            angle = angle * toRADIANS;
    
            var vector:Number3D = Number3D.FORWARD.clone();
    
            updateTransform();
    
            Matrix3D.rotateAxis(transform, vector);
            var m:Matrix3D = Matrix3D.rotationMatrix(vector.x, vector.y, vector.z, angle);
    
            this.transform.copy3x3(Matrix3D.multiply3x3(m ,transform));
    
            this._rotationDirty = true;
        }
    
        // Make the object look at a specific position.
        // @param targetObject Object to look at.
        // @param upAxis The vertical axis of the universe. Normally the positive Y axis.
        public function lookAt(targetObject:Object3D, upAxis:Number3D = null):void
        {
            var position:Number3D = new Number3D(this.x, this.y, this.z);
            var target:Number3D = new Number3D(targetObject.x, targetObject.y, targetObject.z);
    
            var zAxis:Number3D = Number3D.sub(target, position);
            zAxis.normalize();
    
            if (zAxis.modulo > 0.1)
            {
                var xAxis:Number3D = Number3D.cross(zAxis, upAxis || Number3D.UP);
                xAxis.normalize();
    
                var yAxis:Number3D = Number3D.cross(zAxis, xAxis);
                yAxis.normalize();
    
                var look:Matrix3D = this.transform;
    
                look.n11 = xAxis.x;
                look.n21 = xAxis.y;
                look.n31 = xAxis.z;
    
                look.n12 = -yAxis.x;
                look.n22 = -yAxis.y;
                look.n32 = -yAxis.z;
    
                look.n13 = zAxis.x;
                look.n23 = zAxis.y;
                look.n33 = zAxis.z;
    
                this._transformDirty = false;
                this._rotationDirty = true;
                // TODO: Implement scale
            }
            else
            {
                throw new Error("lookAt Error");
            }
        }

        // Copies the transformation information (position, rotation and scale) from another object or Matrix3D.
        // @param reference A Object3D or Matrix3D object to copy the position from.
        public function copyTransform(reference:Object3D):void
        {
            reference.updateTransform();

            var matrix:Matrix3D = reference.transform;
    
            transform.n11 = matrix.n11;   transform.n12 = matrix.n12;
            transform.n13 = matrix.n13;   transform.n14 = matrix.n14;
    
            transform.n21 = matrix.n21;   transform.n22 = matrix.n22;
            transform.n23 = matrix.n23;   transform.n24 = matrix.n24;
    
            transform.n31 = matrix.n31;   transform.n32 = matrix.n32;
            transform.n33 = matrix.n33;   transform.n34 = matrix.n34;
    
            this._transformDirty = false;
            this._rotationDirty = true;
        }

        // [internal-use] Updates the transform Matrix3D with the current rotation and scale values.
        protected function updateTransform():void
        {
            if (!this._transformDirty) 
                return;

            var q:Quaternion = Matrix3D.euler2quaternion(-this._rotationY, -this._rotationZ, this._rotationX); // Swapped
            var m:Matrix3D = Matrix3D.quaternion2matrix(q.x, q.y, q.z, q.w);
            m.n14 = transform.n14;
            m.n24 = transform.n24;
            m.n34 = transform.n34;
            m.scale(_scaleX, _scaleY, _scaleZ);

            this.transform = m;
            this._transformDirty = false;
        }
    
        // Returns a string value representing the three-dimensional position values of the display object instance.
        public function toString():String
        {
            return this.name + ': x:' + Math.round(this.x) + ' y:' + Math.round(this.y) + ' z:' + Math.round(this.z);
        }
    
        private static var toDEGREES:Number = 180 / Math.PI;
        private static var toRADIANS:Number = Math.PI / 180;

        public function tick(time:int):void
        {
        }
    }
}
